#ifndef __ARCHI_HH__
#define __ARCHI_HH__

#include <vector>
#include <map>
#include <stack>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
//#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/sax/SAXParseException.hpp>
//#include <xercesc/dom/DOMErrorHandler.hpp>
#include <xercesc/sax/ErrorHandler.hpp>
#include <xercesc/dom/DOMTreeWalker.hpp>
#include <iostream>
#include <fstream>

#include "c.hh"
#include "pattern.hh" //Graph for representing patterns
#include "tinst.hh"

using namespace xercesc;

class ErrorReport : public ErrorHandler
{

public:

  ErrorReport()
//    fSawErrors(false) 
  {
  }
  ~ErrorReport()
  {
  }

  void warning(const SAXParseException & toCatch);
  void error(const SAXParseException & toCatch);
  void fatalError(const SAXParseException & toCatch);
  void resetErrors() {}
  //{ 
   // fSawErrors = false;
  //}

//  bool getSawErrors() const
 // {
  //  return fSawErrors;
  //}

//private:
 // bool fSawErrors;

};

//Already defined FIXME
typedef std::vector< bool > bit_vector_t;
typedef std::vector< bit_vector_t > reserv_tab_t;

// struct ltstr
// {
//   bool operator()(const char* s1, const char* s2) const
//   {
//     return strcmp(s1, s2) < 0;
//   }
// };


//mapping between op number and list of <pattern *, TargetInstruction *>
typedef std::pair< pattern_t *, TargetInstruction * > pattern_instr_t;
typedef std::map< int, std::vector< pattern_instr_t * > > op_map_t;
typedef std::vector< pattern_instr_t *>::iterator match_iter;
typedef std::map< int, std::vector< std::pair< int, TargetInstruction * > > > transfer_map_t;

class Architecture {

public:

  Architecture(char *);
  ~Architecture();

  void parse(void);
  int get_omega(void);
  char *get_reg_name(int);
  std::pair< match_iter, match_iter > match_start(int);
  int get_reserv_tab_size(void) { return stage_map.size(); }
  std::vector< std::vector< int > > & get_rc_resources(void) { return rc_resources; }
  
  std::pair< bool, std::vector< std::pair< int, TargetInstruction * > > >
  get_transfer_destinations(int _source);

  /** Useful for finding string values directly under
   *  the <architecture>, such as <slotseparator>||</slotseparator>
   * @param key The key
   * @return The defintion
   */
  char *get_defined_string(char *key);
  /** Return the number of functional units of the node
   *  <funits>...</funits> 
   *
   * @return Number of functional units
   */
  int get_funits(void);
  /** Return the cost for activating a functional unit,
   *  specifed as acost attribute for a functional unit.
   *
   *  @parma _fu The functional unit id
   *  @return activation cost
   */
  float get_activation_cost(int _fu);

  /** Return true if the operation _op should be removed from the 
   *  DAG construction.
   *
   * @param _op Operation number
   * @return true if the operation should not appear in the DAG.
   */
  bool remove_op(int _op);
  
  /** Generates ILP data for the ILP formulation.
   *
   * @param _f Output file to which write the data
   * @return void
   */
  void ilp(std::ofstream & _f);

private:

  char *xadml_file_name;
  XercesDOMParser *parser;
  DOMDocument *spec;

  DOMNode *get_sub_tree(const char *, DOMNode *_current_node = NULL);
  void build_pattern_map(void);
  void build_register_map(void);
  void build_residence_map(void);
  void build_transfer_map(void);

  void initialize(void);
  void build_op_map(std::ofstream &);
  int build_target_instruction(DOMNode * _target, 
			       std::vector< TargetInstruction * > & _target_instrs,
			       std::ofstream &);
  void build_format(DOMNode *, char * &);
  void build_constraint(DOMNode *, char * &);

  //Pattern help methods
  void build_pattern_kid(DOMNode *_dom_node, int _kid_nr,
			char *_kid_id, pvertex_id & _v, pattern_t & _pattern);
  void build_pattern_node(DOMNode *_dom_node, pvertex_id & _v, 
			 pattern_t & _pattern);
  void build_tree_pattern(DOMNode *_dom_node, pattern_t & _pattern);

  //Instruction help methods
  void build_reserv_tab(DOMNode *_reserv, reserv_tab_t & _reserv_tab);

  //Help methods op_map
  void build_op_map_instruction(DOMNode *, std::ofstream &);
  void build_op_map_pattern(DOMNode *, std::ofstream &);
  void build_id_rclass_map(DOMNode *ops, 
			   std::map< char *, int, ltstr > & _id_rclass_map,
			   reserv_tab_t *_reserv_tab);
  void check_match(pattern_t *_pattern,
		   std::map< char *, int, ltstr > _id_rclass_map,
		   std::vector< int > & op_numbers);

  //Mapping pattern_name -> pattern (build after parsing)
  typedef std::map< char *, pattern_t *, ltstr > pattern_map_t;
  //residence_name -> vector of registers
  std::map< char *, std::vector< int >, ltstr > residence_map;
  std::map< char *, int, ltstr > residence_index;
  std::vector< std::vector < int > > rc_resources;

  struct reg {
    int size;
    int index;
  };

  //Each register has a unique index (r1, r2, ...)
  std::map< char *, struct reg, ltstr > register_map;
  //Reverse mapping (index->register)
  std::map< int, char * > register_index;

  pattern_map_t pattern_map;
  op_map_t op_map;
  //source rc -> vector of transfers
  transfer_map_t transfer_map;
  pattern_t fake;

  //reservation table
  //Each name in the stage pipline coresponds to 
  //the entry of the reservation table
  //fetch -> 2, i.e. the second column in the reservation table.
  //reservation table size is the number of stage names.

  std::map< char *, int, ltstr > stage_map;
  void build_stage_map(void);

  //Functional units part
  //maps name to entry
  std::map< char *, int, ltstr > funits_map;
  struct fu_data_s {
    char *id;
    int occupation, latency;
    float activation_cost;
  };
  std::vector< fu_data_s > fu_id_map;
  void build_fu_map(void);
  void build_fus(DOMNode *, std::vector< int > &);
  void get_fu_data(DOMNode *, float *, std::vector< int > &);

  //Set of operation to be removed from the DAG
  void build_remove_ops(void);
  std::set< int > remove_ops;

  //Conditions
  void build_target_condition(DOMNode *, std::ofstream &, TargetInstruction *, pattern_t *_pattern = NULL);
  void C_condition(DOMNode *_condition, std::ofstream &, std::stack< int > &, pattern_t *, bool b_first = true);
  void cond(DOMNode *_condition, std::ofstream & _f, 
	    std::stack< std::pair< bool, bool > > & _logic_op,
	    pattern_t *_pattern);

  //  std::ofstream *f;
  bool b_out;
  void start_cond_file(std::ofstream &);
  void end_cond_file(std::ofstream &);
  int cond_index;

  typedef bool (*condition)(std::vector< lcc::Node >);
  void *handle;
  condition *conditions;
  std::string lookup_arg(std::string, pattern_t *);

  std::map< const char *, int, ltstr > ops_map;

  int get_op_number(char *);
  void addop(int op, int type, const char *sz, const char *opname);

  struct data_s {
    std::string target_pattern_id;
    std::vector< std::pair< int, std::pair< int ,int > > > nodes; //node id, <operator_id, out_deg>
    std::vector< std::pair< int, int > > edges; //source, destination
    int latency;
    std::string resource; //use resource
  };

  struct singleton_s {
    std::string instruction_id;
    int op_id;
    int latency;
    std::string resource;
  };

public:
  condition *get_conditions(void) { return conditions; }

};

#endif //__ARCHI_HH__
